#!/usr/bin/python3
# -*- coding: utf-8 -*-
#Page 159
import pylab

pylab.figure(1) # figure 1 を作成
pylab.plot([1,2,3,4], [1,7,3,5]) # figure 1 に作図
pylab.show()

#Page 160-161
pylab.figure(1) # figure 1 を作成
pylab.plot([1,2,3,4], [1,2,3,4]) # figure 1 に作図
pylab.figure(2) # figure 2 を作成
pylab.plot([1,4,2,3], [5,6,7,8]) # figure 2 に作図
pylab.savefig('Figure-Addie') # figure 2 を保存
pylab.figure(1) # figure 1 の作業に戻る
pylab.plot([5,6,10,3]) # figure 1 に再度作図
pylab.savefig('Figure-Jane') # figure 1 に保存


#Page 161
principal = 10000 # 初期投資
interestRate = 0.05
years = 20
values = []
for i in range(years + 1):
    values.append(principal)
    principal += principal*interestRate

pylab.plot(values)
pylab.show()

#Page 162
pylab.title('5% Growth, Compounded Annually')
pylab.xlabel('Years of Compounding')
pylab.ylabel('Value of Principal ($)')

#Page 163
principal = 10000 # 初期投資
interestRate = 0.05
years = 20
values = []
for i in range(years + 1):
    values.append(principal)
    principal += principal*interestRate

pylab.plot(values, linewidth = 30)
pylab.title('5% Growth, Compounded Annually',
            fontsize = 'xx-large')
pylab.xlabel('Years of Compounding', fontsize = 'x-small')
pylab.ylabel('Value of Principal ($)')
pylab.show()

# 線の太さの設定
pylab.rcParams['lines.linewidth'] = 4
# 題名に使われる文字の大きさの設定
pylab.rcParams['axes.titlesize'] = 20
# 軸の名前に使われる文字の大きさの設定
pylab.rcParams['axes.labelsize'] = 20
# x 軸の数字の大きさの設定
pylab.rcParams['xtick.labelsize'] = 16
# y 軸の数字の大きさの設定
pylab.rcParams['ytick.labelsize'] = 16
# x 軸の目盛り幅の設定
pylab.rcParams['xtick.major.size'] = 7
# y 軸の目盛り幅の設定
pylab.rcParams['ytick.major.size'] = 7
# マーカーの大きさの設定
pylab.rcParams['lines.markersize'] = 10

### 注意: プログラムのインポートをしやすいように，
### 上のコードは rcParameterSettings.py にある．

# 以下の関数は 8 章で定義され，本章で使用される．
def findPayment(loan, r, m):
    """loan と r を float 型とし，m を int 型とする
       月割りの金利を r として，借入額 loan の住宅ローンを
       m ヶ月で返済する場合の，毎月の返済額を返す"""
    return loan*((r*(1+r)**m)/((1+r)**m - 1))

#Page 164, Figure 11.1
class Mortgage(object):
    """異なる住宅ローンを扱うための抽象クラス"""
    #    
    def __init__(self, loan, annRate, months):
        """新たな住宅ローンを生成する"""
        self.loan = loan
        self.rate = annRate/12.0
        self.months = months
        self.paid = [0.0] 
        self.owed = [loan]
        self.payment = findPayment(loan, self.rate, months)
        self.legend = None # 住宅ローンの種類（サブクラスで用いる）
    #            
    def makePayment(self):
        """返済を行う"""
        self.paid.append(self.payment)
        reduction = self.payment - self.owed[-1]*self.rate
        self.owed.append(self.owed[-1] - reduction)
    #            
    def getTotalPaid(self):
        """これまで支払った総額を返す"""
        return sum(self.paid)
    #        
    def __str__(self):
        return self.legend
    #    
    def plotPayments(self, style):
        pylab.plot(self.paid[1:], style, label = self.legend)
    #        
    def plotBalance(self, style):
        pylab.plot(self.owed, style, label = self.legend)
    #            
    def plotTotPd(self, style):
        """これまでに支払った総額をプロットする"""
        totPd = [self.paid[0]]
        for i in range(1, len(self.paid)):
            totPd.append(totPd[-1] + self.paid[i])
        pylab.plot(totPd, style, label = self.legend)
    #        
    def plotNet(self, style):
        """住宅ローンの総費用の近似値を時系列的にプロットする
           具体的には，総支払額からローンの返済によって得た純資産額を
           減じたものをプロットする"""
        totPd = [self.paid[0]]
        for i in range(1, len(self.paid)):
            totPd.append(totPd[-1] + self.paid[i])
        # 支払いによって得る純資産額とは，ローンの総額からローンの残高を減じたもの
        equityAcquired = pylab.array([self.loan]*len(self.owed))
        equityAcquired = equityAcquired - pylab.array(self.owed)
        net = pylab.array(totPd) - equityAcquired
        pylab.plot(net, style, label = self.legend)

#Page 165-166
a1 = pylab.array([1, 2, 4])
print( 'a1 =', a1)
a2 = a1*2
print( 'a2 =', a2)
print( 'a1 + 3 =', a1 + 3)
print( '3 - a1 =', 3 - a1)
print( 'a1 - a2 =', a1 - a2)
print( 'a1*a2 =', a1*a2)

#Page 166, Figure 11.2
class Fixed(Mortgage):
    def __init__(self, loan, r, months):
        Mortgage.__init__(self, loan, r, months)
        self.legend = 'Fixed, ' + str(r*100) + '%' 

class FixedWithPts(Mortgage):
    def __init__(self, loan, r, months, pts):
        Mortgage.__init__(self, loan, r, months)
        self.pts = pts
        self.paid = [loan*(pts/100.0)]
        self.legend = 'Fixed, ' + str(r*100) + '%, '\
                   + str(pts) + ' points'

class TwoRate(Mortgage):
    def __init__(self, loan, r, months, teaserRate, teaserMonths):
        Mortgage.__init__(self, loan, teaserRate, months)
        self.teaserMonths = teaserMonths
        self.teaserRate = teaserRate
        self.nextRate = r/12.0
        self.legend = str(teaserRate*100)\
                      + '% for ' + str(self.teaserMonths)\
                      + ' months, then ' + str(r*100) + '%'
    #
    def makePayment(self):
        if len(self.paid) == self.teaserMonths + 1:
            self.rate = self.nextRate
            self.payment = findPayment(self.owed[-1], self.rate,
                                       self.months - self.teaserMonths)
        Mortgage.makePayment(self)

#Page 167-168, Figure 11.3
def plotMortgages(morts, amt):
    styles = ['b-', 'b-.', 'b:']
    # 図を指定する際，番号の変わりに名前を用いる
    payments = 0
    cost = 1
    balance = 2
    netCost = 3
    pylab.figure(payments)
    pylab.title('Monthly Payments of Different $' + str(amt)
                + ' Mortgages')
    pylab.xlabel('Months')
    pylab.ylabel('Monthly Payments')
    pylab.figure(cost)
    pylab.title('Cash Outlay of Different $' + str(amt) + ' Mortgages')
    pylab.xlabel('Months')
    pylab.ylabel('Total Payments')
    pylab.figure(balance)
    pylab.title('Balance Remaining of $' + str(amt) + ' Mortgages')
    pylab.xlabel('Months')
    pylab.ylabel('Remaining Loan Balance of $')
    pylab.figure(netCost)
    pylab.title('Net Cost of $' + str(amt) + ' Mortgages')
    pylab.xlabel('Months')
    pylab.ylabel('Payments - Equity $')
    for i in range(len(morts)):
        pylab.figure(payments)
        morts[i].plotPayments(styles[i])
        pylab.figure(cost)
        morts[i].plotTotPd(styles[i])
        pylab.figure(balance)
        morts[i].plotBalance(styles[i])
        pylab.figure(netCost)
        morts[i].plotNet(styles[i])
    pylab.figure(payments)
    pylab.legend(loc = 'upper center')
    pylab.figure(cost)
    pylab.legend(loc = 'best')
    pylab.figure(balance)
    pylab.legend(loc = 'best') 

def compareMortgages(amt, years, fixedRate, pts, ptsRate,
                    varRate1, varRate2, varMonths):
    totMonths = years*12
    fixed1 = Fixed(amt, fixedRate, totMonths)
    fixed2 = FixedWithPts(amt, ptsRate, totMonths, pts)
    twoRate = TwoRate(amt, varRate2, totMonths, varRate1, varMonths)
    morts = [fixed1, fixed2, twoRate]
    for m in range(totMonths):
        for mort in morts:
            mort.makePayment()
    plotMortgages(morts, amt)

#Page 168
compareMortgages(amt=200000, years=30, fixedRate=0.07,
                 pts = 3.25, ptsRate=0.05,
                 varRate1=0.045, varRate2=0.095, varMonths=48)
